@@ -23,6 +23,7 @@
 #include <config.h>
 #endif
 
+#include <errno.h>
 #include <string.h>
 
 #include <gdbus.h>
@@ -33,17 +34,18 @@ struct connman_network {
 	struct connman_element element;
 	enum connman_network_type type;
 	enum connman_network_protocol protocol;
+	connman_bool_t associating;
 	connman_bool_t secondary;
 	connman_bool_t available;
 	connman_bool_t connected;
-	connman_bool_t remember;
-	connman_uint8_t priority;
 	connman_uint8_t strength;
+	connman_uint16_t frequency;
 	char *identifier;
 	char *address;
 	char *name;
 	char *node;
 	char *group;
+	struct connman_ipconfig *ipconfig;
 
 	struct connman_network_driver *driver;
 	void *driver_data;
@@ -56,6 +58,7 @@ struct connman_network {
 		void *ssid;
 		int ssid_len;
 		char *mode;
+		unsigned short channel;
 		char *security;
 		char *passphrase;
 	} wifi;
@@ -120,23 +123,17 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		connman_dbus_dict_append_variant(&dict, "Name",
 					DBUS_TYPE_STRING, &network->name);
 
-	connman_dbus_dict_append_variant(&dict, "Available",
-				DBUS_TYPE_BOOLEAN, &network->available);
-
 	connman_dbus_dict_append_variant(&dict, "Connected",
 				DBUS_TYPE_BOOLEAN, &network->connected);
 
-	connman_dbus_dict_append_variant(&dict, "Remember",
-				DBUS_TYPE_BOOLEAN, &network->remember);
-
-	if (network->priority > 0)
-		connman_dbus_dict_append_variant(&dict, "Priority",
-					DBUS_TYPE_BYTE, &network->priority);
-
 	if (network->strength > 0)
 		connman_dbus_dict_append_variant(&dict, "Strength",
 					DBUS_TYPE_BYTE, &network->strength);
 
+	if (network->frequency > 0)
+		connman_dbus_dict_append_variant(&dict, "Frequency",
+					DBUS_TYPE_UINT16, &network->frequency);
+
 	if (network->wifi.ssid != NULL && network->wifi.ssid_len > 0)
 		connman_dbus_dict_append_array(&dict, "WiFi.SSID",
 				DBUS_TYPE_BYTE, &network->wifi.ssid,
@@ -146,6 +143,10 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		connman_dbus_dict_append_variant(&dict, "WiFi.Mode",
 				DBUS_TYPE_STRING, &network->wifi.mode);
 
+	if (network->wifi.channel > 0)
+		connman_dbus_dict_append_variant(&dict, "WiFi.Channel",
+				DBUS_TYPE_UINT16, &network->wifi.channel);
+
 	if (network->wifi.security != NULL)
 		connman_dbus_dict_append_variant(&dict, "WiFi.Security",
 				DBUS_TYPE_STRING, &network->wifi.security);
@@ -184,19 +185,7 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 	type = dbus_message_iter_get_arg_type(&value);
 
-	if (g_str_equal(name, "Remember") == TRUE) {
-		connman_bool_t remember;
-
-		if (type != DBUS_TYPE_BOOLEAN)
-			return __connman_error_invalid_arguments(msg);
-
-		dbus_message_iter_get_basic(&value, &remember);
-
-		if (network->remember == remember)
-			return __connman_error_invalid_arguments(msg);
-
-		network->remember = remember;
-	} else if (g_str_equal(name, "WiFi.Passphrase") == TRUE) {
+	if (g_str_equal(name, "WiFi.Passphrase") == TRUE) {
 		const char *passphrase;
 
 		if (type != DBUS_TYPE_STRING)
@@ -210,16 +199,15 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		g_free(network->wifi.passphrase);
 		network->wifi.passphrase = g_strdup(passphrase);
-	} else if (g_str_equal(name, "Priority") == TRUE) {
-		connman_uint8_t priority;
+	} else if (g_str_has_prefix(name, "IPv4.") == TRUE) {
+		int err;
 
-		if (type != DBUS_TYPE_BYTE)
-			return __connman_error_invalid_arguments(msg);
-
-		dbus_message_iter_get_basic(&value, &priority);
-
-		network->priority = priority;
-	}
+		err = __connman_ipconfig_set_ipv4(network->ipconfig,
+							name + 5, &value);
+		if (err < 0)
+			return __connman_error_failed(msg, -err);
+	} else
+		return __connman_error_invalid_property(msg);
 
 	__connman_storage_save_network(network);
 
@@ -239,7 +227,7 @@ static DBusMessage *do_connect(DBusConnection *conn,
 		return __connman_error_permission_denied(msg);
 
 	if (network->connected == TRUE)
-		return __connman_error_failed(msg);
+		return __connman_error_failed(msg, EALREADY);
 
 	if (network->driver && network->driver->connect) {
 		enum connman_device_mode mode;
@@ -250,7 +238,7 @@ static DBusMessage *do_connect(DBusConnection *conn,
 
 		err = network->driver->connect(network);
 		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg);
+			return __connman_error_failed(msg, -err);
 	} else
 		network->connected = TRUE;
 
@@ -270,7 +258,7 @@ static DBusMessage *do_disconnect(DBusConnection *conn,
 		return __connman_error_permission_denied(msg);
 
 	if (network->connected == FALSE)
-		return __connman_error_failed(msg);
+		return __connman_error_failed(msg, EINVAL);
 
 	connman_element_unregister_children(&network->element);
 
@@ -279,7 +267,7 @@ static DBusMessage *do_disconnect(DBusConnection *conn,
 	if (network->driver && network->driver->disconnect) {
 		err = network->driver->disconnect(network);
 		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg);
+			return __connman_error_failed(msg, -err);
 	} else
 		network->connected = FALSE;
 
@@ -440,6 +428,8 @@ static void network_destruct(struct connman_element *element)
 	g_free(network->name);
 	g_free(network->address);
 	g_free(network->identifier);
+
+	connman_ipconfig_unref(network->ipconfig);
 }
 
 /**
@@ -491,6 +481,12 @@ struct connman_network *connman_network_create(const char *identifier,
 	network->secondary  = FALSE;
 	network->identifier = g_strdup(identifier);
 
+	network->ipconfig = connman_ipconfig_create();
+	if (network->ipconfig == NULL) {
+		connman_network_unref(network);
+		return NULL;
+	}
+
 	return network;
 }
 
@@ -603,8 +599,34 @@ void connman_network_set_protocol(struct connman_network *network,
 void connman_network_set_group(struct connman_network *network,
 							const char *group)
 {
-	g_free(network->group);
+	if (network->secondary == TRUE)
+		return;
+
+	if (g_strcmp0(network->group, group) == 0)
+		return;
+
+	switch (network->type) {
+	case CONNMAN_NETWORK_TYPE_UNKNOWN:
+	case CONNMAN_NETWORK_TYPE_VENDOR:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+	case CONNMAN_NETWORK_TYPE_HSO:
+		return;
+	case CONNMAN_NETWORK_TYPE_WIFI:
+	case CONNMAN_NETWORK_TYPE_WIMAX:
+		break;
+	}
+
+	if (network->group != NULL) {
+		__connman_profile_remove_network(network);
+
+		g_free(network->group);
+	}
+
 	network->group = g_strdup(group);
+
+	if (network->group != NULL)
+		__connman_profile_add_network(network);
 }
 
 const char *__connman_network_get_group(struct connman_network *network)
@@ -630,10 +652,6 @@ const char *__connman_network_get_ident(struct connman_network *network)
 int connman_network_set_available(struct connman_network *network,
 						connman_bool_t available)
 {
-	DBusMessage *signal;
-	DBusMessageIter entry, value;
-	const char *key = "Available";
-
 	DBG("network %p available %d", network, available);
 
 	if (network->available == available)
@@ -641,31 +659,6 @@ int connman_network_set_available(struct connman_network *network,
 
 	network->available = available;
 
-	if (network->registered == FALSE)
-		return 0;
-
-	if (network->connected == TRUE)
-		return 0;
-
-	if (network->remember == FALSE)
-		return 0;
-
-	signal = dbus_message_new_signal(network->element.path,
-				CONNMAN_NETWORK_INTERFACE, "PropertyChanged");
-	if (signal == NULL)
-		return 0;
-
-	dbus_message_iter_init_append(signal, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
-	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN, &available);
-	dbus_message_iter_close_container(&entry, &value);
-
-	g_dbus_send_message(connection, signal);
-
 	return 0;
 }
 
@@ -680,9 +673,40 @@ connman_bool_t connman_network_get_available(struct connman_network *network)
 	return network->available;
 }
 
+/**
+ * connman_network_set_associating:
+ * @network: network structure
+ * @associating: associating state
+ *
+ * Change associating state of network
+ */
+int connman_network_set_associating(struct connman_network *network,
+						connman_bool_t associating)
+{
+	DBG("network %p associating %d", network, associating);
+
+	if (network->associating == associating)
+		return -EALREADY;
+
+	network->associating = associating;
+
+	if (associating == TRUE) {
+		struct connman_service *service;
+
+		service = __connman_service_lookup_from_network(network);
+		__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_ASSOCIATION);
+	}
+
+	return 0;
+}
+
 static gboolean set_connected(gpointer user_data)
 {
 	struct connman_network *network = user_data;
+	struct connman_service *service;
+
+	service = __connman_service_lookup_from_network(network);
 
 	if (network->connected == TRUE) {
 		struct connman_element *element;
@@ -713,8 +737,14 @@ static gboolean set_connected(gpointer user_data)
 			if (connman_element_register(element,
 						&network->element) < 0)
 				connman_element_unref(element);
+
+			__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_CONFIGURATION);
 		}
 	} else {
+		__connman_service_indicate_state(service,
+						CONNMAN_SERVICE_STATE_IDLE);
+
 		connman_element_unregister_children(&network->element);
 
 		__connman_device_set_network(network->device, NULL);
@@ -746,6 +776,9 @@ int connman_network_set_connected(struct connman_network *network,
 
 	network->connected = connected;
 
+	if (connected == TRUE)
+		network->associating = FALSE;
+
 	if (network->registered == FALSE) {
 		g_idle_add(set_connected, network);
 		return 0;
@@ -784,91 +817,133 @@ connman_bool_t connman_network_get_connected(struct connman_network *network)
 }
 
 /**
- * connman_network_set_remember:
- * @network: network structure
- * @remember: remember state
- *
- * Change remember state of network (known networks)
- */
-int connman_network_set_remember(struct connman_network *network,
-						connman_bool_t remember)
-{
-	DBusMessage *signal;
-	DBusMessageIter entry, value;
-	const char *key = "Remember";
-
-	DBG("network %p remember %d", network, remember);
-
-	if (network->remember == remember)
-		return -EALREADY;
-
-	network->remember = remember;
-
-	if (network->registered == FALSE)
-		return 0;
-
-	signal = dbus_message_new_signal(network->element.path,
-				CONNMAN_NETWORK_INTERFACE, "PropertyChanged");
-	if (signal == NULL)
-		return 0;
-
-	dbus_message_iter_init_append(signal, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
-	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN, &remember);
-	dbus_message_iter_close_container(&entry, &value);
-
-	g_dbus_send_message(connection, signal);
-
-	return 0;
-}
-
-/**
- * connman_network_get_remember:
- * @network: network structure
- *
- * Get network remember setting
- */
-connman_bool_t connman_network_get_remember(struct connman_network *network)
-{
-	return network->remember;
-}
-
-/**
- * connman_network_connect:
+ * __connman_network_connect:
  * @network: network structure
  *
  * Connect network
  */
-int connman_network_connect(struct connman_network *network)
+int __connman_network_connect(struct connman_network *network)
 {
+	int err;
+
+	DBG("network %p", network);
+
 	if (network->connected == TRUE)
 		return -EALREADY;
 
-	if (network->driver && network->driver->connect)
-		return network->driver->connect(network);
+	if (network->driver == NULL)
+		return -EUNATCH;
 
-	network->connected = TRUE;
+	if (network->driver->connect == NULL)
+		return -ENOSYS;
 
-	return 0;
+	__connman_device_disconnect(network->device);
+
+	err = network->driver->connect(network);
+	if (err == 0) {
+		network->connected = TRUE;
+		set_connected(network);
+	}
+
+	return err;
 }
 
+/**
+ * __connman_network_disconnect:
+ * @network: network structure
+ *
+ * Disconnect network
+ */
 int __connman_network_disconnect(struct connman_network *network)
 {
+	int err;
+
+	DBG("network %p", network);
+
 	if (network->connected == FALSE)
 		return -ENOTCONN;
 
-	connman_element_unregister_children(&network->element);
+	if (network->driver == NULL)
+		return -EUNATCH;
 
-	if (network->driver && network->driver->disconnect)
-		return network->driver->disconnect(network);
+	if (network->driver->disconnect == NULL)
+		return -ENOSYS;
 
-	network->connected = FALSE;
+	err = network->driver->disconnect(network);
+	if (err == 0) {
+		network->connected = FALSE;
+		set_connected(network);
+	}
 
-	return 0;
+	return err;
+}
+
+/**
+ * connman_network_set_address:
+ * @network: network structure
+ * @address: binary address value
+ * @size: binary address length
+ *
+ * Set unique address value for network
+ */
+int connman_network_set_address(struct connman_network *network,
+				const void *address, unsigned int size)
+{
+	const unsigned char *addr_octet = address;
+	char *str;
+
+	DBG("network %p size %d", network, size);
+
+	if (size != 6)
+		return -EINVAL;
+
+	str = g_strdup_printf("%02X:%02X:%02X:%02X:%02X:%02X",
+				addr_octet[0], addr_octet[1], addr_octet[2],
+				addr_octet[3], addr_octet[4], addr_octet[5]);
+	if (str == NULL)
+		return -ENOMEM;
+
+	g_free(network->address);
+	network->address = str;
+
+	return connman_element_set_string(&network->element,
+						"Address", network->address);
+}
+
+/**
+ * connman_network_set_name:
+ * @network: network structure
+ * @name: name value
+ *
+ * Set display name value for network
+ */
+int connman_network_set_name(struct connman_network *network,
+							const char *name)
+{
+	DBG("network %p name %s", network, name);
+
+	g_free(network->name);
+	network->name = g_strdup(name);
+
+	return connman_element_set_string(&network->element, "Name", name);
+}
+
+/**
+ * connman_network_set_strength:
+ * @network: network structure
+ * @strength: strength value
+ *
+ * Set signal strength value for network
+ */
+int connman_network_set_strength(struct connman_network *network,
+						connman_uint8_t strength)
+{
+	DBG("network %p strengh %d", network, strength);
+
+	network->strength = strength;
+
+	return connman_element_set_uint8(&network->element,
+						"Strength", strength);
 }
 
 /**
@@ -884,12 +959,12 @@ int connman_network_set_string(struct connman_network *network,
 {
 	DBG("network %p key %s value %s", network, key, value);
 
+	if (g_strcmp0(key, "Name") == 0)
+		return connman_network_set_name(network, value);
+
 	if (g_str_equal(key, "Address") == TRUE) {
 		g_free(network->address);
 		network->address = g_strdup(value);
-	} else if (g_str_equal(key, "Name") == TRUE) {
-		g_free(network->name);
-		network->name = g_strdup(value);
 	} else if (g_str_equal(key, "Node") == TRUE) {
 		g_free(network->node);
 		network->node = g_strdup(value);
@@ -948,10 +1023,8 @@ int connman_network_set_uint8(struct connman_network *network,
 {
 	DBG("network %p key %s value %d", network, key, value);
 
-	if (g_str_equal(key, "Priority") == TRUE)
-		network->priority = value;
-	else if (g_str_equal(key, "Strength") == TRUE)
-		network->strength = value;
+	if (g_strcmp0(key, "Strength") == 0)
+		return connman_network_set_strength(network, value);
 
 	return connman_element_set_uint8(&network->element, key, value);
 }
@@ -968,14 +1041,53 @@ connman_uint8_t connman_network_get_uint8(struct connman_network *network,
 {
 	DBG("network %p key %s", network, key);
 
-	if (g_str_equal(key, "Priority") == TRUE)
-		return network->priority;
-	else if (g_str_equal(key, "Strength") == TRUE)
+	if (g_str_equal(key, "Strength") == TRUE)
 		return network->strength;
 
 	return connman_element_get_uint8(&network->element, key);
 }
 
+/**
+ * connman_network_set_uint16:
+ * @network: network structure
+ * @key: unique identifier
+ * @value: integer value
+ *
+ * Set integer value for specific key
+ */
+int connman_network_set_uint16(struct connman_network *network,
+				const char *key, connman_uint16_t value)
+{
+	DBG("network %p key %s value %d", network, key, value);
+
+	if (g_str_equal(key, "Frequency") == TRUE)
+		network->frequency = value;
+	else if (g_str_equal(key, "WiFi.Channel") == TRUE)
+		network->wifi.channel = value;
+
+	return -EINVAL;
+}
+
+/**
+ * connman_network_get_uint16:
+ * @network: network structure
+ * @key: unique identifier
+ *
+ * Get integer value for specific key
+ */
+connman_uint16_t connman_network_get_uint16(struct connman_network *network,
+							const char *key)
+{
+	DBG("network %p key %s", network, key);
+
+	if (g_str_equal(key, "Frequency") == TRUE)
+		return network->frequency;
+	else if (g_str_equal(key, "WiFi.Channel") == TRUE)
+		return network->wifi.channel;
+
+	return 0;
+}
+
 /**
  * connman_network_set_blob:
  * @network: network structure
@@ -990,6 +1102,9 @@ int connman_network_set_blob(struct connman_network *network,
 {
 	DBG("network %p key %s size %d", network, key, size);
 
+	if (g_strcmp0(key, "Address") == 0)
+		return connman_network_set_address(network, data, size);
+
 	if (g_str_equal(key, "WiFi.SSID") == TRUE) {
 		g_free(network->wifi.ssid);
 		network->wifi.ssid = g_try_malloc(size);
@@ -1121,7 +1236,7 @@ static int network_probe(struct connman_element *element)
 		break;
 	case CONNMAN_NETWORK_TYPE_WIFI:
 	case CONNMAN_NETWORK_TYPE_WIMAX:
-		if (network->secondary == FALSE)
+		if (network->group != NULL && network->secondary == FALSE)
 			__connman_profile_add_network(network);
 		break;
 	}
@@ -1150,8 +1265,12 @@ static void network_remove(struct connman_element *element)
 		break;
 	case CONNMAN_NETWORK_TYPE_WIFI:
 	case CONNMAN_NETWORK_TYPE_WIMAX:
-		if (network->secondary == FALSE)
+		if (network->group != NULL && network->secondary == FALSE) {
 			__connman_profile_remove_network(network);
+
+			g_free(network->group);
+			network->group = NULL;
+		}
 		break;
 	}
 
@@ -1197,156 +1316,12 @@ static struct connman_driver network_driver = {
 	.change		= network_change,
 };
 
-static int network_init(struct connman_device *device)
-{
-	DBG("device %p", device);
-
-	return 0;
-}
-
-static int network_load(struct connman_network *network)
-{
-	GKeyFile *keyfile;
-	gchar *pathname, *data = NULL;
-	gsize length;
-	const char *name;
-	char *str;
-	int val;
-
-	DBG("network %p", network);
-
-	name = connman_device_get_name(network->device);
-	if (name == NULL)
-		return -EINVAL;
-
-	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR, name);
-	if (pathname == NULL)
-		return -ENOMEM;
-
-	keyfile = g_key_file_new();
-
-	if (g_file_get_contents(pathname, &data, &length, NULL) == FALSE) {
-		g_free(pathname);
-		return -ENOENT;
-	}
-
-	g_free(pathname);
-
-	if (g_key_file_load_from_data(keyfile, data, length,
-							0, NULL) == FALSE) {
-		g_free(data);
-		return -EILSEQ;
-	}
-
-	g_free(data);
-
-	network->remember = g_key_file_get_boolean(keyfile,
-					network->identifier, "Remember", NULL);
-
-	val = g_key_file_get_integer(keyfile, network->identifier,
-							"Priority", NULL);
-	if (val > 0)
-		network->priority = val;
-
-	str = g_key_file_get_string(keyfile,
-				network->identifier, "WiFi.Security", NULL);
-	if (str != NULL) {
-		g_free(network->wifi.security);
-		network->wifi.security = str;
-	}
-
-	str = g_key_file_get_string(keyfile,
-				network->identifier, "WiFi.Passphrase", NULL);
-	if (str != NULL) {
-		g_free(network->wifi.passphrase);
-		network->wifi.passphrase = str;
-	}
-
-	g_key_file_free(keyfile);
-
-	return 0;
-}
-
-static int network_save(struct connman_network *network)
-{
-	GKeyFile *keyfile;
-	gchar *pathname, *data = NULL;
-	gsize length;
-	const char *name;
-
-	DBG("network %p", network);
-
-	name = connman_device_get_name(network->device);
-	if (name == NULL)
-		return -EINVAL;
-
-	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR, name);
-	if (pathname == NULL)
-		return -ENOMEM;
-
-	keyfile = g_key_file_new();
-
-	if (g_file_get_contents(pathname, &data, &length, NULL) == FALSE)
-		goto update;
-
-	if (length > 0) {
-		if (g_key_file_load_from_data(keyfile, data, length,
-							0, NULL) == FALSE)
-			goto done;
-	}
-
-	g_free(data);
-
-update:
-	g_key_file_set_boolean(keyfile, network->identifier,
-					"Remember", network->remember);
-
-	if (network->priority > 0)
-		g_key_file_set_integer(keyfile, network->identifier,
-						"Priority", network->priority);
-
-	if (network->remember == TRUE || network->connected == TRUE) {
-		if (network->wifi.security != NULL)
-			g_key_file_set_string(keyfile, network->identifier,
-				"WiFi.Security", network->wifi.security);
-
-		if (network->wifi.passphrase != NULL)
-			g_key_file_set_string(keyfile, network->identifier,
-				"WiFi.Passphrase", network->wifi.passphrase);
-	}
-
-	data = g_key_file_to_data(keyfile, &length, NULL);
-
-	if (g_file_set_contents(pathname, data, length, NULL) == FALSE)
-		connman_error("Failed to store network information");
-
-done:
-	g_free(data);
-
-	g_key_file_free(keyfile);
-
-	g_free(pathname);
-
-	return 0;
-}
-
-static struct connman_storage network_storage = {
-	.name		= "network",
-	.priority	= CONNMAN_STORAGE_PRIORITY_LOW,
-	.network_init	= network_init,
-	.network_load	= network_load,
-	.network_save	= network_save,
-};
-
 int __connman_network_init(void)
 {
 	DBG("");
 
 	connection = connman_dbus_get_connection();
 
-	if (connman_storage_register(&network_storage) < 0)
-		connman_error("Failed to register network storage");
-
 	return connman_driver_register(&network_driver);
 }
 
@@ -1356,7 +1331,5 @@ void __connman_network_cleanup(void)
 
 	connman_driver_unregister(&network_driver);
 
-	connman_storage_unregister(&network_storage);
-
 	dbus_connection_unref(connection);
 }
